/* 
 *   ccglue - cscope,ctags glue: Process ctags and cscope output to
 *        produce cross-reference table (for use with Vim CCTree plugin)
 *        and more ...
 *   Copyright (C) April, 2011,  Hari Rangarajan 
 *   Contact: hariranga@users.sourceforge.net

 *   "ccglue" is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.

 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.

 *   You should have received a copy of the GNU General Public License
 *   along with "ccglue".  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <glib.h>
#include <stdio.h>
#include "sym_mgr.h"
#include "digraph.h"
#include "debug.h"
#include <string.h>
#include "seqfile.h"
//#include <gprintf.h>

struct sym_table {
	uint32		monotonic_id;
	GHashTable* hash_names;
	GArray*     array_sym;
};

struct sym_entry {
	uint32	uid;	
	GString *uid_str;
	GString *n;
	GString *p;
	GString *c;
};

RC_t sym_table_init (sym_table_t** table)
{
	*table = (sym_table_t *) malloc(sizeof(sym_table_t));
	if (*table == NULL) {
		return RC_FAILURE;
	}
	if (((*table)->hash_names = g_hash_table_new(g_str_hash, g_str_equal)) 
					== NULL) {
		free(*table);
		return RC_FAILURE;
	}
	if (((*table)->array_sym = g_array_new(FALSE, FALSE,
                        sizeof(sym_entry_t *))) == NULL) {
		/* need to free */
		return RC_FAILURE;
	}
	(*table)->monotonic_id = 0;
	return RC_SUCCESS;
}

static void sym_table_sym_name_data_free_iterator (gpointer key,
        gpointer value, gpointer user_data)
{
    free((uchar *) key);
    sym_entry_destroy((sym_entry_t *) value);
    free(value);
}

static void sym_table_internal_data_destroy (sym_table_t* table)
{
    g_hash_table_foreach(table->hash_names, 
            sym_table_sym_name_data_free_iterator, NULL);
}

void sym_table_destroy (sym_table_t* table)
{
        sym_table_internal_data_destroy(table);
	g_array_free(table->array_sym, TRUE);
	g_hash_table_destroy(table->hash_names);
        free(table);
}

uint32 sym_table_get_new_id (sym_table_t* table)
{
	return table->monotonic_id++;
}

RC_t sym_entry_create (sym_entry_t **sym_entry)
{
	*sym_entry = (sym_entry_t *) malloc(sizeof(sym_entry_t));
	if (*sym_entry == NULL) {
		return RC_FAILURE;
	}
	/* allocate strings */
	(*sym_entry)->n = g_string_new(NULL);
	(*sym_entry)->p = g_string_new(NULL);
	(*sym_entry)->c = g_string_new(NULL);
	(*sym_entry)->uid_str = g_string_new(NULL);

	return RC_SUCCESS;
}

RC_t sym_entry_destroy (sym_entry_t *sym_entry)
{
    if (sym_entry) {
        g_string_free(sym_entry->n, TRUE);
        g_string_free(sym_entry->p, TRUE);
        g_string_free(sym_entry->c, TRUE);
        g_string_free(sym_entry->uid_str, TRUE);
    }
}

gint sym_table_array_entry_cmp (gconstpointer a, gconstpointer b)
{
	sym_entry_t	*sym1 = (sym_entry_t *) (*(uint32 *)a);
	sym_entry_t	*sym2 = (sym_entry_t *) (*(uint32 *)b);

	return strcmp(sym1->uid_str->str, sym2->uid_str->str);
}

sym_entry_t* sym_table_add (sym_table_t *table, uchar *sym_name)
{
	sym_entry_t *sym_entry;

	if (sym_entry_create(&sym_entry) != RC_SUCCESS) {
		return NULL;
	}
	g_string_printf(sym_entry->n, "%s", sym_name);
	sym_entry->uid = sym_table_get_new_id(table);
	g_string_printf(sym_entry->uid_str, "%ld", sym_entry->uid);
	
	//DEBUG_PRINTF("Adding %s on %ld\n", sym_entry->n->str, sym_entry->uid);

	g_hash_table_insert(table->hash_names, (gpointer)g_strdup(sym_name),
					(gpointer)sym_entry);

	/* g_list_insert_sorted(table->list_sym, (gpointer)sym_entry, 
           sym_table_list_insert_cmp);
	*/
        g_array_append_val(table->array_sym, sym_entry);
	return sym_entry;
}

sym_entry_t* sym_table_add_ext (sym_table_t *table, uchar *sym_name)
{
    sym_entry_t *sym_entry = sym_table_lookup(table, sym_name);
    if (sym_entry == NULL) {
        //DEBUG_PRINTF(DBG_VERBOSE, "Could not find %s \n", sym_name);
        sym_entry = sym_table_add(table, sym_name);
    } 
    return sym_entry;
}

sym_entry_t* sym_table_lookup (sym_table_t *table, uchar *sym_name)
{
	sym_entry_t *sym_entry;

	sym_entry = (sym_entry_t *) g_hash_table_lookup(table->hash_names, 
			(gpointer) sym_name);
	return sym_entry;
}

void sym_entry_mark_p (sym_entry_t* self, sym_entry_t *p) 
{
	g_string_append_printf(self->p, "%ld,", p->uid);
}

void sym_entry_mark_c (sym_entry_t* self, sym_entry_t *c) 
{
	g_string_append_printf(self->c, "%ld,", c->uid);
}
							
void sym_table_mark_xref (sym_entry_t* in_func, sym_entry_t *ref_func) 
{
	sym_entry_mark_c(in_func, ref_func);
	sym_entry_mark_p(ref_func, in_func);
}
	

#if 0
void sym_table_dump_hash_iterator (gpointer key, gpointer value, gpointer user_data) 
{
	sym_entry_t *sym_entry = (sym_entry_t *) value;
	digraph_uncompress_string_with_map(sym_entry->n->str, tmp, 
						(digraph_map_t *) user_data);
	printf("%ld#%s\t\t/^\\$/;\" ", sym_entry->uid, tmp);
	digraph_compress_string_with_map(sym_entry->c->str, tmp, digraph_get_numeric_map());
	printf("\tc:%s", tmp);
	digraph_compress_string_with_map(sym_entry->p->str, tmp, digraph_get_numeric_map());
	printf("\tp:%s\n", tmp);
}
void sym_table_dump_hash (sym_table_t *table)
{
	g_hash_table_foreach(table->hash_names, (GHFunc)sym_table_dump_hash_iterator, 
						digraph_get_letter_map());
}
#endif

void sym_table_dump_array_iterator (seq_file_t *seq_file, 
                                        sym_entry_t *sym_entry) 
{
    char tmp[1024];
    uchar *tstr;

    sprintf(tmp, "%s#", sym_entry->uid_str->str);
    seq_file_write_string(seq_file, tmp);
    digraph_uncompress_string_with_map_write_to_seqfile(sym_entry->n->str, 
            seq_file, digraph_get_letter_map());
    
    sprintf(tmp, "\t\t/^\\$/;\" ", sym_entry->uid_str->str);
    seq_file_write_string(seq_file, tmp);

    tstr = "\tc:";
    seq_file_write_string(seq_file, tstr);

    tstr = &sym_entry->c->str[0];
    if (digraph_compress_string_with_map_write_to_seqfile(tstr, seq_file,
                digraph_get_numeric_map()) == RC_FAILURE) {
        perror("Compress seqfile failed");
        return;
    }

    tstr = "\tp:";
    seq_file_write_string(seq_file, tstr);
    tstr = &sym_entry->p->str[0];
    if (digraph_compress_string_with_map_write_to_seqfile(tstr, seq_file,
                digraph_get_numeric_map()) == RC_FAILURE) {
        perror("Compress seqfile failed");
        return;
    }
    seq_file_write_char(seq_file, '\n');
} 

void sym_table_dump_array_fp (seq_file_t *seq_file, sym_table_t *table)
{
    int i;

	g_array_sort(table->array_sym, sym_table_array_entry_cmp);
	for (i = 0; i < table->array_sym->len; i++) {
			sym_table_dump_array_iterator(seq_file, 
                        g_array_index(table->array_sym, sym_entry_t*, i));
	}
}

void sym_table_write_xref_tag_header (seq_file_t *seq_file)
{
    uchar *tstr;

    tstr = "!_TAG_FILE_FORMAT\t2\t/extended format; --format=1 will not"
                " append ;\" to lines/\n";
    seq_file_write_buffer(seq_file, tstr, strlen(tstr));

    tstr = "!_TAG_FILE_SORTED\t1\t/0=unsorted, 1=sorted, 2=foldcase/\n";
    seq_file_write_buffer(seq_file, tstr, strlen(tstr));
    tstr = "!_TAG_PROGRAM_NAME\t\tccglue for use with CCTree (Vim plugin)//\n";
    seq_file_write_buffer(seq_file, tstr, strlen(tstr));
    tstr = "!_TAG_PROGRAM_URL\thttp://vim.sourceforge.net/scrip"
                "ts/script.php?script_id=2368\t/site/\n";
    seq_file_write_buffer(seq_file, tstr, strlen(tstr));
}

void sym_table_write_xref_tag_file (uchar* fname, sym_table_t *table)
{
    seq_file_t  *seq_file;

    if (seq_file_writer_create(&seq_file, fname) == RC_SUCCESS) {
        sym_table_write_xref_tag_header(seq_file);
        sym_table_dump_array_fp(seq_file, table);
    }        
    seq_file_destroy(seq_file);
}
